import {desiredHeight, desiredWidth, getScreen} from "./engine/internal/screen.ts";
import {BG_OUTER} from "./colors.ts";
import {D, I} from "./engine/public.ts";
import {IGame, Point, Size} from "./engine/datatypes.ts";
import {HuntMode} from "./huntmode.ts";

class MenuCamera {
  // measured in whole screens
  position: Point;
  target: Point;

  constructor({position, target}: {position: Point, target: Point}) {
    this.position = position;
    this.target = target;
  }

  update() {
    let adjust = (x0: number, x1: number) => {
      if (Math.abs(x1 - x0) < 0.01) { return x1; }
      return (x0 * 8 + x1 * 2) / 10;
    }
    this.position = new Point(
      adjust(this.position.x, this.target.x),
      adjust(this.position.y, this.target.y),
    );
  }
}

type GameState = "Gameplay" | "Thralls";

function getScreenLocation(state: GameState): Point {
  if (state === "Gameplay") {
    return new Point(0, 0);
  }
  if (state === "Thralls") {
    return new Point(0, 1);
  }

  throw `invalid state: ${state}`
}

export class Game implements IGame {
  camera: MenuCamera;
  state: GameState;
  huntMode: HuntMode;
  frame: number;

  constructor() {
    this.camera = new MenuCamera({
      position: new Point(0, 0),
      target: new Point(0, 0),
    });
    this.state = "Gameplay";

    this.huntMode = HuntMode.generate({depth: 1});
    this.frame = 0;
  }

  update() {
    this.frame += 1;
    if (I.isKeyPressed("w")) {
      this.state = "Gameplay"
    }
    if (I.isKeyPressed("s")) {
      this.state = "Thralls"
    }

    this.camera.target = getScreenLocation(this.state);
    D.camera = new Point(
      D.size.w * this.camera.position.x,
      D.size.h * this.camera.position.y,
    )
    this.camera.update();

    // state-specific updates
    this.updateGameplay();
  }

  draw() {
    // draw screen background
    let oldCamera = D.camera;
    D.camera = new Point(0, 0);
    D.fillRect(new Point(0, 0), D.size, BG_OUTER);
    D.camera = oldCamera;

    this.drawGameplay();

    // we draw all states at once and pan between them
    // mainFont.drawText({ctx: ctx, text: "You have been given a gift.", x: 0, y: 0})
    let mouse = I.mousePosition?.offset(D.camera);
    if (mouse != null) {
      D.invertRect(mouse.offset(new Point(-1, -1)), new Size(3, 3))
    }
  }

  getPaneRegionForGameState(gameState: GameState) {
    let screen = getScreen();
    let {w, h} = screen.size;
    let overallScreenLocation = getScreenLocation(gameState);

    let bigPaneX = overallScreenLocation.x * w;
    let bigPaneY = overallScreenLocation.y * h;
    let bigPaneW = w;
    let bigPaneH = h;

    let smallPaneW = desiredWidth;
    let smallPaneH = desiredHeight;
    let smallPaneX = Math.floor(bigPaneX + (bigPaneW - smallPaneW) / 2)
    let smallPaneY = Math.floor(bigPaneY + (bigPaneH - smallPaneH) / 2)

    return {
      big: {
        position: new Point(bigPaneX, bigPaneY),
        size: new Size(bigPaneW, bigPaneH),
      },
      small: {
        position: new Point(smallPaneX, smallPaneY),
        size: new Size(smallPaneW, smallPaneH),
      }
    }
  }

  #withHuntModeCamera(cb: () => void) {
    let region = this.getPaneRegionForGameState("Gameplay")

    D.withCamera(D.camera.offset(region.small.position.negate()), cb)
  }

  updateGameplay() {
    this.#withHuntModeCamera(() => {
      this.huntMode.update();
    });
  }

  drawGameplay() {
    this.#withHuntModeCamera(() => {
      this.huntMode.draw();
    });
  }
}


export let game = new Game();